import json
import time
from websockets.sync.client import connect as ws_connect
from websockets.exceptions import ConnectionClosedError

_default_host = "127.0.0.1"
_default_port = 30008

_n_retries = 10
_initial_retry = 0.01  # seconds
_shutdown_delay = 0.1  # seconds


def connect(host=_default_host, port=_default_port, verbose=False):
    return DSMQClientSideConnection(host, port, verbose=verbose)


class DSMQClientSideConnection:
    def __init__(self, host, port, verbose=False):
        self.uri = f"ws://{host}:{port}"
        self.verbose = verbose
        if self.verbose:
            print(f"Connecting to dsmq server at {self.uri}")
        for i_retry in range(_n_retries):
            try:
                self.websocket = ws_connect(self.uri)
                break
            except ConnectionRefusedError:
                self.websocket = None
                # Exponential backoff
                # Wait twice as long each time before trying again.
                time.sleep(_initial_retry * 2**i_retry)
                if self.verbose:
                    print("    ...trying again")

        if self.websocket is None:
            raise ConnectionRefusedError("Could not connect to dsmq server.")

        self.time_of_last_request = time.time()

    def get(self, topic):
        msg = {"action": "get", "topic": topic}
        self.websocket.send(json.dumps(msg))
        try:
            msg_text = self.websocket.recv()
        except ConnectionClosedError:
            self.close()

        msg = json.loads(msg_text)
        return msg["message"]

    def get_wait(self, topic):
        """
        A variant of `get()` that retries a few times until it gets
        a non-empty message. Adjust `_n_tries` and `_initial_retry`
        to change how persistent it will be.
        """
        for i_retry in range(_n_retries):
            message = self.get(topic)
            if message != "":
                return message
            time.sleep(_initial_retry * 2**i_retry)
        return message

    def put(self, topic, msg_body):
        msg_dict = {"action": "put", "topic": topic, "message": msg_body}
        self.websocket.send(json.dumps(msg_dict))

    def shutdown_server(self):
        msg_dict = {"action": "shutdown", "topic": ""}
        self.websocket.send(json.dumps(msg_dict))
        # Give the server time to wind down
        time.sleep(_shutdown_delay)

    def close(self):
        self.websocket.close()
        # Give the websocket time to wind down
        time.sleep(_shutdown_delay)



import time
from dsmq.client import connect


def run(host="127.0.0.1", port=30008, n_iter=1000):
    mq = connect(host=host, port=port)

    for i in range(n_iter):
        time.sleep(1)
        msg = mq.get("greetings")
        print(f"client received {msg}")


if __name__ == "__main__":
    run()



import multiprocessing as mp
from dsmq.server import serve
import dsmq.example_get_client
import dsmq.example_put_client

HOST = "127.0.0.1"
PORT = 25252


def test_server_with_clients():
    p_server = mp.Process(target=serve, args=(HOST, PORT))
    p_server.start()

    p_putter = mp.Process(target=dsmq.example_put_client.run, args=(HOST, PORT, 20))
    p_getter = mp.Process(target=dsmq.example_get_client.run, args=(HOST, PORT, 20))

    p_putter.start()
    p_getter.start()


if __name__ == "__main__":
    test_server_with_clients()



import json
import os
import sqlite3
import sys
from threading import Thread
import time
from websockets.sync.server import serve as ws_serve
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK

_default_host = "127.0.0.1"
_default_port = 30008
_n_retries = 20
_first_retry = 0.005  # seconds
_time_to_live = 60.0  # seconds

# _db_name = ":memory:"
_db_name = "file::memory:?cache=shared"
# May occasionally create files with this name.
# https://sqlite.org/inmemorydb.html
# "...parts of a temporary database might be flushed to disk if the
# database becomes large or if SQLite comes under memory pressure."

# Make this global so it's easy to share
dsmq_server = None


def serve(host=_default_host, port=_default_port, verbose=False):
    """
    For best results, start this running in its own process and walk away.
    """
    # Cleanup temp files.
    # Under some condition
    # (which I haven't yet been able to pin down)
    # a file is generated with the db name.
    # If it is not removed, it gets
    # treated as a SQLite db on disk,
    # which dramatically slows it down,
    # especially the way it's used here for
    # rapid-fire one-item reads and writes.
    filenames = os.listdir()
    for filename in filenames:
        if filename[: len(_db_name)] == _db_name:
            os.remove(filename)

    sqlite_conn = sqlite3.connect(_db_name)
    cursor = sqlite_conn.cursor()

    # Tweak the connection to make it faster
    # and keep long-term latency more predictable.
    # cursor.execute("PRAGMA journal_mode = OFF")
    cursor.execute("PRAGMA journal_mode = WAL")
    # cursor.execute("PRAGMA synchronous = OFF")
    # cursor.execute("PRAGMA secure_delete = OFF")

    cursor.execute("""
CREATE TABLE IF NOT EXISTS messages (timestamp DOUBLE, topic TEXT, message TEXT)
    """)

    # Making this global in scope is a way to make it available
    # to the shutdown operation. It's an awkward construction,
    # and a method of last resort. (If you stumble across this and
    # figure out something more elegant, please submit a PR!
    # or send it to me at brohrer@gmail.com,
    global dsmq_server

    # dsmq_server = ws_serve(request_handler, host, port)
    for i_retry in range(_n_retries):
        try:
            with ws_serve(request_handler, host, port) as dsmq_server:
                dsmq_server.serve_forever()

            if verbose:
                print()
                print(f"Server started at {host} on port {port}.")
                print("Waiting for clients...")

            break

        except OSError:
            # Catch the case where the address is already in use
            if verbose:
                print()
                if i_retry < _n_retries - 1:
                    print(f"Couldn't start dsmq server on {host} on port {port}.")
                    print(f"    Trying again ({i_retry}) ...")
                else:
                    print()
                    print(f"Failed to start dsmq server on {host} on port {port}.")
                    print()
                    raise

            wait_time = _first_retry * 2**i_retry
            time.sleep(wait_time)

    sqlite_conn.close()


def request_handler(websocket):
    sqlite_conn = sqlite3.connect(_db_name)
    cursor = sqlite_conn.cursor()

    client_creation_time = time.time()
    last_read_times = {}
    time_of_last_purge = time.time()

    try:
        for msg_text in websocket:
            msg = json.loads(msg_text)
            topic = msg["topic"]
            timestamp = time.time()

            if msg["action"] == "put":
                msg["timestamp"] = timestamp

                # This block allows for multiple retries if the database
                # is busy.
                for i_retry in range(_n_retries):
                    try:
                        cursor.execute(
                            """
                            INSERT INTO messages (timestamp, topic, message)
                            VALUES (:timestamp, :topic, :message)
                            """,
                            (msg),
                        )
                        sqlite_conn.commit()
                    except sqlite3.OperationalError:
                        wait_time = _first_retry * 2**i_retry
                        time.sleep(wait_time)
                        continue
                    break

            elif msg["action"] == "get":
                try:
                    last_read_time = last_read_times[topic]
                except KeyError:
                    last_read_times[topic] = client_creation_time
                    last_read_time = last_read_times[topic]
                msg["last_read_time"] = last_read_time

                # This block allows for multiple retries if the database
                # is busy.
                for i_retry in range(_n_retries):
                    try:
                        cursor.execute(
                            """
                            SELECT message,
                            timestamp
                            FROM messages,
                            (
                            SELECT MIN(timestamp) AS min_time
                            FROM messages
                            WHERE topic = :topic
                            AND timestamp > :last_read_time
                            ) a
                            WHERE topic = :topic
                            AND timestamp = a.min_time
                            """,
                            msg,
                        )
                    except sqlite3.OperationalError:
                        wait_time = _first_retry * 2**i_retry
                        time.sleep(wait_time)
                        continue
                    break

                try:
                    result = cursor.fetchall()[0]
                    message = result[0]
                    timestamp = result[1]
                    last_read_times[topic] = timestamp
                except IndexError:
                    # Handle the case where no results are returned
                    message = ""

                websocket.send(json.dumps({"message": message}))
            elif msg["action"] == "shutdown":
                # Run this from a separate thread to prevent deadlock
                global dsmq_server

                def shutdown_gracefully(server_to_shutdown):
                    server_to_shutdown.shutdown()

                    filenames = os.listdir()
                    for filename in filenames:
                        if filename[: len(_db_name)] == _db_name:
                            try:
                                os.remove(filename)
                            except FileNotFoundError:
                                pass

                Thread(target=shutdown_gracefully, args=(dsmq_server,)).start()
                break
            else:
                raise RuntimeWarning(
                    "dsmq client action must either be 'put', 'get', or 'shutdown'"
                )

            # Periodically clean out messages from the queue that are
            # past their sell buy date.
            # This operation is pretty fast. I clock it at 12 us on my machine.
            if time.time() - time_of_last_purge > _time_to_live:
                try:
                    cursor.execute(
                        """
                        DELETE FROM messages
                        WHERE timestamp < :time_threshold
                        """,
                        {"time_threshold": time_of_last_purge},
                    )
                    sqlite_conn.commit()
                    time_of_last_purge = time.time()
                except sqlite3.OperationalError:
                    # Database may be locked. Try again next time.
                    pass
    except (ConnectionClosedError, ConnectionClosedOK):
        # Something happened on the other end and this handler
        # is no longer needed.
        pass

    sqlite_conn.close()


if __name__ == "__main__":
    if len(sys.argv) == 3:
        host = sys.argv[1]
        port = int(sys.argv[2])
        serve(host=host, port=port)
    elif len(sys.argv) == 2:
        host = sys.argv[1]
        serve(host=host)
    elif len(sys.argv) == 1:
        serve()
    else:
        print(
            """
Try one of these:
$ python3 server.py

$ python3 server.py 127.0.0.1

$ python3 server.py 127.0.0.1 25853

"""
        )



import multiprocessing as mp
import time
from dsmq.server import serve
from dsmq.client import connect

# spawn is the default method on macOS
# mp.set_start_method('spawn')

host = "127.0.0.1"
port = 30303

_short_pause = 0.001
_pause = 0.01
_long_pause = 0.1
_very_long_pause = 2.0


def test_client_server():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    mq = connect(host, port)
    read_completes = False
    write_completes = False

    n_iter = 11
    for i in range(n_iter):
        mq.put("test", f"msg_{i}")
        write_completes = True

    for i in range(n_iter):
        msg = mq.get("test")
        read_completes = True

    assert msg
    assert write_completes
    assert read_completes

    mq.shutdown_server()
    mq.close()

    closed = False
    try:
        mq = connect(host, port)
    except ConnectionRefusedError as e:
        print(e)
        closed = True
    assert closed


def test_write_one_read_one():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client = connect(host, port)
    read_client = connect(host, port)

    write_client.put("test", "test_msg")

    # It takes a moment for the write to complete
    time.sleep(_pause)
    msg = read_client.get("test")

    assert msg == "test_msg"

    write_client.shutdown_server()
    write_client.close()
    read_client.close()


def test_get_wait():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client = connect(host, port)
    read_client = connect(host, port)

    write_client.put("test", "test_msg")

    msg = read_client.get_wait("test")

    assert msg == "test_msg"

    write_client.shutdown_server()
    write_client.close()
    read_client.close()


def test_multitopics():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client = connect(host, port)
    read_client = connect(host, port)

    write_client.put("test_A", "test_msg_A")
    write_client.put("test_B", "test_msg_B")

    msg_A = read_client.get_wait("test_A")
    msg_B = read_client.get_wait("test_B")

    assert msg_A == "test_msg_A"
    assert msg_B == "test_msg_B"

    write_client.shutdown_server()
    write_client.close()
    read_client.close()


def test_client_history_cutoff():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client = connect(host, port)
    write_client.put("test", "test_msg")
    time.sleep(_pause)

    read_client = connect(host, port)
    msg = read_client.get("test")

    assert msg == ""

    write_client.shutdown_server()
    write_client.close()
    read_client.close()


def test_two_write_clients():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client_A = connect(host, port)
    write_client_B = connect(host, port)
    read_client = connect(host, port)

    write_client_A.put("test", "test_msg_A")
    # Wait briefly, to ensure the order of writes
    time.sleep(_pause)
    write_client_B.put("test", "test_msg_B")
    msg_A = read_client.get_wait("test")
    msg_B = read_client.get_wait("test")

    assert msg_A == "test_msg_A"
    assert msg_B == "test_msg_B"

    write_client_A.shutdown_server()
    write_client_A.close()
    write_client_B.close()
    read_client.close()


def test_two_read_clients():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client = connect(host, port)
    read_client_A = connect(host, port)
    read_client_B = connect(host, port)

    write_client.put("test", "test_msg")
    msg_A = read_client_A.get_wait("test")
    msg_B = read_client_B.get_wait("test")

    assert msg_A == "test_msg"
    assert msg_B == "test_msg"

    write_client.shutdown_server()
    write_client.close()
    read_client_A.close()
    read_client_B.close()


def speed_write(stop_flag):
    fast_write_client = connect(host, port)
    while True:
        fast_write_client.put("speed_test", "speed")
        if stop_flag.is_set():
            break
    fast_write_client.close()


def test_speed_writing():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client = connect(host, port)
    read_client = connect(host, port)

    stop_flag = mp.Event()
    p_speed_write = mp.Process(target=speed_write, args=(stop_flag,))
    p_speed_write.start()
    time.sleep(_pause)

    # time_a = time.time()
    write_client.put("test", "test_msg")
    # time_b = time.time()
    msg = read_client.get_wait("test")
    # time_c = time.time()

    # write_time = int((time_b - time_a) * 1e6)
    # read_time = int((time_c - time_b) * 1e6)
    # print(f"write time: {write_time} us,  read time: {read_time} us")

    assert msg == "test_msg"

    stop_flag.set()
    p_speed_write.join()
    read_client.close()
    write_client.shutdown_server()
    write_client.close()


def speed_read(stop_flag):
    fast_read_client = connect(host, port)
    while True:
        fast_read_client.get("speed_test")
        if stop_flag.is_set():
            break
    fast_read_client.close()


def test_speed_reading():
    p_server = mp.Process(target=serve, args=(host, port))
    p_server.start()
    write_client = connect(host, port)
    read_client = connect(host, port)

    stop_flag = mp.Event()
    p_speed_read = mp.Process(target=speed_read, args=(stop_flag,))
    p_speed_read.start()

    # time_a = time.time()
    write_client.put("test", "test_msg")
    # time_b = time.time()
    msg = read_client.get_wait("test")
    # time_c = time.time()

    # write_time = int((time_b - time_a) * 1e6)
    # read_time = int((time_c - time_b) * 1e6)
    # print(f"write time: {write_time} us,  read time: {read_time} us")

    assert msg == "test_msg"

    stop_flag.set()
    p_speed_read.join()
    p_speed_read.kill()
    read_client.close()
    write_client.shutdown_server()
    # time.sleep(_pause)
    write_client.close()



import multiprocessing as mp
import time

from dsmq.server import serve
from dsmq.client import connect

host = "127.0.0.1"
port = 30303
verbose = False

_pause = 0.01
_very_long_pause = 1.0

_n_iter = int(1e3)
_n_long_char = int(1e4)

_short_msg = "q"
_long_msg = str(["q"] * _n_long_char)

_test_topic = "test"


def main():
    print()
    print("dsmq timing measurements")

    time_short_writes()
    time_long_writes()
    time_empty_reads()
    time_short_reads()
    time_long_reads()


def time_short_writes():
    condition = "short write"

    duration, duration_close = time_writes(msg=_short_msg, n_iter=1)

    print()
    print(f"Time for first {condition}  [including closing]")
    print(f"        {int(duration)} μs  [{int(duration_close)} μs]")

    avg_duration, avg_duration_close = time_writes(msg=_short_msg, n_iter=_n_iter)

    print(f"Average time for a {condition}  [including closing]")
    print(f"        {int(avg_duration)} μs  [{int(avg_duration_close)} μs]")


def time_long_writes():
    duration, duration_close = time_writes(msg=_long_msg, n_iter=1)

    condition = "long write"
    print()
    print(f"Time for first {condition}  [including closing]")
    print(f"        {int(duration)} μs  [{int(duration_close)} μs]")

    avg_duration, avg_duration_close = time_writes(msg=_long_msg, n_iter=_n_iter)

    condition = f"long write ({_n_long_char} characters)"
    print(f"Average time for a {condition}  [including closing]")
    print(f"        {int(avg_duration)} μs  [{int(avg_duration_close)} μs]")

    condition = "long write (per 1000 characters)"
    print(f"Average time for a {condition}  [including closing]")
    print(
        f"        {int(1000 * avg_duration / _n_long_char)} μs  "
        + f"[{int(1000 * avg_duration_close / _n_long_char)}] μs"
    )


def time_writes(msg="message", n_iter=1):
    p_server = mp.Process(target=serve, args=(host, port, verbose))
    p_server.start()
    time.sleep(_pause)
    write_client = connect(host, port)

    start_time = time.time()
    for _ in range(n_iter):
        write_client.put(_test_topic, msg)
    avg_duration = 1e6 * (time.time() - start_time) / n_iter  # microseconds

    write_client.shutdown_server()
    write_client.close()

    p_server.join(_very_long_pause)
    if p_server.is_alive():
        print("    Doing a hard shutdown on mq server")
        p_server.kill()
    avg_duration_close = 1e6 * (time.time() - start_time) / n_iter  # microseconds

    return avg_duration, avg_duration_close


def time_empty_reads():
    condition = "empty read"

    duration, duration_close = time_reads(msg=None, n_iter=1)

    print()
    print(f"Time for first {condition}  [including closing]")
    print(f"        {int(duration)} μs  [{int(duration_close)} μs]")

    avg_duration, avg_duration_close = time_reads(msg=None, n_iter=_n_iter)

    print(f"Average time for a {condition}  [including closing]")
    print(f"        {int(avg_duration)} μs  [{int(avg_duration_close)} μs]")


def time_short_reads():
    condition = "short read"

    duration, duration_close = time_reads(msg=_short_msg, n_iter=1)

    print()
    print(f"Time for first {condition}  [including closing]")
    print(f"        {int(duration)} μs  [{int(duration_close)} μs]")

    avg_duration, avg_duration_close = time_reads(msg=_short_msg, n_iter=_n_iter)

    print(f"Average time for a {condition}  [including closing]")
    print(f"        {int(avg_duration)} μs  [{int(avg_duration_close)} μs]")


def time_long_reads():
    condition = f"long read ({_n_long_char} characters)"

    duration, duration_close = time_reads(msg=_long_msg, n_iter=1)

    print()
    print(f"Time for first {condition}  [including closing]")
    print(f"        {int(duration)} μs  [{int(duration_close)} μs]")

    avg_duration, avg_duration_close = time_reads(msg=_long_msg, n_iter=_n_iter)

    print(f"Average time for a {condition}  [including closing]")
    print(f"        {int(avg_duration)} μs  [{int(avg_duration_close)} μs]")

    condition = "long read (per 1000 characters)"
    print(f"Average time for a {condition}  [including closing]")
    print(
        f"        {int(1000 * avg_duration / _n_long_char)} μs  "
        + f"[{int(1000 * avg_duration_close / _n_long_char)}] μs"
    )


def time_reads(msg=None, n_iter=1):
    p_server = mp.Process(target=serve, args=(host, port, verbose))
    p_server.start()
    time.sleep(_pause)
    # write_client = connect(host, port)
    read_client = connect(host, port)

    if msg is not None:
        for _ in range(n_iter):
            read_client.put(_test_topic, msg)

    start_time = time.time()
    for _ in range(n_iter):
        msg = read_client.get(_test_topic)

    avg_duration = 1e6 * (time.time() - start_time) / n_iter  # microseconds

    read_client.shutdown_server()
    # write_client.close()
    read_client.close()

    p_server.join(_very_long_pause)

    if p_server.is_alive():
        print("    Doing a hard shutdown on mq server")
        p_server.kill()

    avg_duration_close = 1e6 * (time.time() - start_time) / n_iter  # microseconds

    return avg_duration, avg_duration_close


if __name__ == "__main__":
    main()
